/***************************************************************************
 *                                                                         *
 *                         Powersave Daemon                                *
 *                                                                         *
 *          Copyright (C) 2004,2005 SUSE Linux Products GmbH               *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2 of the License, or (at you   *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it will be useful, but     *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA                  *
 *                                                                         *
 ***************************************************************************/

#include "config.h"
#include <stdio.h>
#include <errno.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <unistd.h>

#include "powerlib.h"
#include "powerlib_local.h"

#define ACPI_BATTERY_DIR  "/proc/acpi/battery"

/* needed to evaluate remaining minutes of battery power
   if present rate is not available (for broken ACPI systems) */
static capacity_snapshot rem_capacity[10] = { {0, 0}, };
static int current_snapshot = 0;

int getBatteriesInfo(BatteryGeneral * bat)
{
	int test;

	if (bat == NULL) {
		errno = EINVAL;
		return -2;
	}
	bat->remPerc = -1;
	bat->remMins = -1;
	bat->charging_state = CHARG_STATE_UNKNOWN;
	bat->rem_charging_mins = -1;

	test = checkACPI();
	if (test == ACPI) {
		if (get_ACPI_General_Battery_Info(bat) < 0) {
			pDebug(DBG_INFO, "Could not get battery info for ACPI subsystem\n");
			return UNKNOWN;
		} else {
			return 1;
		}
	} else if (test == APM) {
		if (get_APM_General_Battery_Info(bat) < 0) {
			pDebug(DBG_INFO, "Could not get battery info for APM subsystem\n");
			return UNKNOWN;
		} else {
			return 1;
		}
	} else {
		return UNKNOWN;
	}
}

int get_ACPI_General_Battery_Info(BatteryGeneral * bat)
{

	int x, present = 0;
	BatteryDetailed bd;
	int remaining_capacity, last_full_capacity, present_rate;
	present_rate = 0;
	remaining_capacity = 0;
	last_full_capacity = 0;
	int cmp = 0;
	float rem_battery_sum = 0;
	int divisor = 0;

	/* check whether module has been loaded and device is available */
	present = check_ACPI_dir(ACPI_BATTERY_DIR);
	if (present < 0)
		/* returns either NO_DEVICE_ERROR or NO_MODULE_ERROR */
		return present;

	/* iterate existing batteries */
	for (x = 0; (getBatteryDetailedInfo(x, &bd) == 0) && x < MAX_BATTERIES; x++) {
		if (bd.present == PRESENT_YES) {
			present = 1;
			if (bd.last_full_capacity > 0) {
				/* last remaining capacity might be greater than last full capacity */
				last_full_capacity +=
				    (bd.remaining_capacity >
				     bd.last_full_capacity) ? bd.remaining_capacity : bd.last_full_capacity;
				/* on buggy batteries remaining capacity might be greater than last_full/design capacity */
				remaining_capacity += bd.remaining_capacity;
				present_rate += bd.present_rate;
			}
			/* keep charging info over multiple batteries */
			bat->charging_state = (bat->charging_state | bd.charging_state);
		}
	}
	if (!present) {
		pDebug(DBG_INFO, "No batteries available.\n");
		return 0;
	}
	// cannot calc anything
	if (remaining_capacity <= 0) {
		return 1;
	} else {

		/* calc remMins ********************************** */
		if (getACAdapterStatus() == AC_OFFLINE) {
			// calc remMins from present_rate and remaining_capacity
			if (present_rate > 0) {
				bat->remMins = (int)roundf(((float)(remaining_capacity) / present_rate) * 60.0);
			}
			// calc remMins depending on used capacity (stored from
			// previous remaining capacities
			else {
				/* previous and current remaining capacity are equal.
				   This happens if measurements are done too often.
				   Use the previous previous measurement then */
				/* value storing the last calulated amount of remaining time */
				static int previous_remMins = -1;
				if (rem_capacity[current_snapshot].remaining_capacity == remaining_capacity) {
					pDebug(DBG_INFO, "charge level unchanged");
					/* bat->remMins is set to -1 in getBatteriesInfo. So we have to
					 * assign the last measurement. Otherwise on systems were battery
					 * discharging is too slow, we would never calculate */
					bat->remMins = previous_remMins;
				} else {
					current_snapshot++;
					if (current_snapshot > 9)
						current_snapshot = 0;
					rem_capacity[current_snapshot].measure_point = time(NULL);
					rem_capacity[current_snapshot].remaining_capacity = remaining_capacity;
					for (int x = current_snapshot - 1, k = 0; k < 9; x--, k++) {
						if (x < 0)
							cmp = x + 10;
						else
							cmp = x;
						/* check whether measurement is valid */
						if (rem_capacity[cmp].measure_point == 0
						    || rem_capacity[current_snapshot].measure_point <
						    rem_capacity[cmp].measure_point
						    || rem_capacity[current_snapshot].remaining_capacity >
						    rem_capacity[cmp].remaining_capacity)
							break;
						/* cmp. with last measure, with the one before until all ten
						   (if available) are cmp. with current rem. capacity.
						   Sum them up and divide later */
						rem_battery_sum +=
						    (float)rem_capacity[current_snapshot].remaining_capacity *
						    (rem_capacity[current_snapshot].measure_point -
						     rem_capacity[cmp].measure_point)
						    /
						    ((rem_capacity[cmp].remaining_capacity -
						      rem_capacity[current_snapshot].remaining_capacity) * 60);
						divisor++;
					}
					/* only calc when we have more than one measurement */
					if (divisor > 1) {
						bat->remMins = rem_battery_sum / divisor;
						/* store the remaining time for the next run were remaining capacity
						   is equal to this one */
						previous_remMins = bat->remMins;
					}
				}
			}
		}
		/* calc remMins ********************************** */

		/* calc remaining charging time until battery is fully charged again
		   if AC online ******* */

		else if (last_full_capacity > 0) {
			if (present_rate > 0) {
				bat->rem_charging_mins = (int)roundf(((float)(last_full_capacity - remaining_capacity)
								      / present_rate) * 60.0);
			}
			/* estimate remaing charging time from increasing remaining capacity,
			   as it is also done for rem minutes (see above)
			 */
			else {
				/* previous and current remaining capacity are equal.
				   This happens if measurements are done too often.
				   Use the previous previous measurement then */
				if (rem_capacity[current_snapshot].remaining_capacity == remaining_capacity) {
					pDebug(DBG_INFO, "charge level unchanged");
				} else {
					current_snapshot++;
					if (current_snapshot > 9)
						current_snapshot = 0;
					rem_capacity[current_snapshot].measure_point = time(NULL);
					rem_capacity[current_snapshot].remaining_capacity = remaining_capacity;
					for (int x = current_snapshot - 1, k = 0; k < 9; x--, k++) {
						if (x < 0)
							cmp = x + 10;
						else
							cmp = x;
						/* check whether measurement is valid */
						if (rem_capacity[cmp].measure_point == 0
						    || rem_capacity[current_snapshot].measure_point <
						    rem_capacity[cmp].measure_point
						    || rem_capacity[current_snapshot].remaining_capacity >
						    rem_capacity[cmp].remaining_capacity)
							break;
						/* cmp. with last measure, with the one before until all ten
						   (if available) are cmp. with current rem. capacity.
						   Sum them up and divide later */
						rem_battery_sum +=
						    (float)(rem_capacity[current_snapshot].remaining_capacity -
							    last_full_capacity)
						    * (rem_capacity[current_snapshot].measure_point -
						       rem_capacity[cmp].measure_point)
						    /
						    ((rem_capacity[current_snapshot].remaining_capacity -
						      rem_capacity[cmp].remaining_capacity)
						     * 60);
						divisor++;
					}
					/* only calc when we have more than one measurement */
					if (divisor > 1) {
						bat->remMins = rem_battery_sum / divisor;
					}
				}
			}
		}
		/* calc remaining chargin time if AC online ****** */

		/* calc remPerc ********************************** */
		if (last_full_capacity > 0) {
			bat->remPerc = (int)roundf(((float)remaining_capacity / (float)last_full_capacity) * 100.0);
			if (bat->remPerc > 150 || bat->remPerc < 0) {
				pDebug(DBG_INFO, "Wrong calculated battery remaining "
				       "Percent capacity value: %d, set to 100", bat->remPerc);
				bat->remPerc = 100;
			} else if (bat->remPerc > 100)
				bat->remPerc = 100;
		}
		/* calc remPerc ********************************** */
	}
	return 1;
}

int get_APM_General_Battery_Info(BatteryGeneral * bg)
{
	struct apm_info apmInfo;
	int t;

	t = read_apm_info(&apmInfo);
	if (t == -2) {
		pDebug(DBG_DIAG, "Old apm style not suppported");
		return -1;
	} else if (t < 0) {
		pDebug(DBG_DIAG, "apm unsuppported, this may not happen, apm support already tested");
		return -1;
	}
	bg->remPerc = apmInfo.battery_percentage;
	if (apmInfo.battery_status == 5) {
		bg->charging_state = CHARG_STATE_CHARGING;
		/*              bg->capacity_state = STATE_OK; */
	} else
		bg->charging_state = CHARG_STATE_DISCHARGING;

	/* there is no warning state in APM, only low and critical ! */
	/* but I do not need it, because I define it myself! */
	/* inside the daemon ...

	   if (apmInfo.battery_status == 3) {
	   bg->capacity_state = STATE_LOW;
	   }
	   else if (apmInfo.battery_status == 4) {
	   bg->capacity_state = STATE_CRIT;
	   }
	 */
	/* no remaining time available, probably on AC power */
	if (apmInfo.battery_time == -1)
		bg->remMins = UNKNOWN;
	/* remaining time in minutes */
	else if (apmInfo.using_minutes) {
		bg->remMins = apmInfo.battery_time;
	}
	/* remaining time in seconds */
	else {
		bg->remMins = apmInfo.battery_time / 60;
	}
	/* no battery flag set ? */
	if (apmInfo.battery_flags & 0x80)
		return -1;
	else
		return 1;
}

/* only for ACPI!!!
return x < 0 on error
returns 0 if alarm has already sub-ceded(alarm will still be set)
returns positive value if alarm has been set successfully
*/

int setBatteryAlarm(int percent)
{
	int x, max_present_bats;
	BatteryDetailed *bd[MAX_BATTERIES];
	/* divide_factor: factor on which all batteries are above the wanted percent (statistically) */
	int remaining_capacity_sum, last_full_capacity_sum, rem_Perc_all;
	float divide_factor = 0.0;
	remaining_capacity_sum = 0;
	last_full_capacity_sum = 0;
	int alarm_value = 0;
	char file[MAX_FILE_PATH + 1] = "";
	char dev_name[MAX_FILE_PATH + 1] = "";
	char verify[MAX_LINE_SIZE + 1] = "";
	char temp[MAX_LINE_SIZE + 1] = "";
	int ret = 1;
	FILE *fp;

	if (checkACPI() != ACPI) {
		return NO_ACPI_ERROR;
	}

	/* check whether module has been loaded and device is available */
	ret = check_ACPI_dir(ACPI_BATTERY_DIR);
	if (ret < 0)
		/* returns either NO_DEVICE_ERROR or NO_MODULE_ERROR */
		return ret;

	if (percent <= 0 || percent > 100) {
		pDebug(DBG_DIAG, "illegal percent value for battery alarm: %d", percent);
		return -1; //UNKNOWN
	}

	for (x = 0; x < MAX_BATTERIES; x++) {
		bd[x] = (BatteryDetailed *) malloc(sizeof(BatteryDetailed));
	}
	/* iterate present battery devices. Don't iterate on existing devices: */
	/* Imagine BAT0 -> not present, BAT1 -> present */
	for (x = 0; (getBatteryDetailedInfo(x, bd[x]) == 0) && x < MAX_BATTERIES; x++) {
		/* I could also get general bat info again, but this would */
		/*result in reading out battery state one more time ... */

		if (bd[x]->present == PRESENT_YES) {
			if (bd[x]->remaining_capacity > 0) {
				if (bd[x]->last_full_capacity > 0) {
					last_full_capacity_sum += bd[x]->last_full_capacity;
					remaining_capacity_sum += bd[x]->remaining_capacity;
				}
			}
		}
	}
	max_present_bats = x;
	/* calculate remaining percent of the capacity of all sumerized batteries ** */
	if (last_full_capacity_sum > 0 && remaining_capacity_sum > 0) {
		rem_Perc_all = (int)roundf(((float)remaining_capacity_sum / (float)last_full_capacity_sum) * 100.0);
		if (rem_Perc_all > 140 || rem_Perc_all < 0) {
			pDebug(DBG_DIAG, "bogus value %d%% remaining, set to 100", rem_Perc_all);
			rem_Perc_all = 100;
		} else if (rem_Perc_all > 100)
			rem_Perc_all = 100;
	} else {
		pDebug(DBG_INFO, "reading last_full and remaining_capacity failed.");
		goto ERROR;
	}
	/*****************************************************************************/
	if (percent > 0)
		divide_factor = (float)rem_Perc_all / percent;
	pDebug(DBG_INFO, "set alarms: rem_Perc_all: %d, want alarm at: %d, divisor: %f", rem_Perc_all, percent, divide_factor);
	for (x = 0; x < max_present_bats && x < MAX_BATTERIES; x++) {
		if (bd[x]->present != PRESENT_YES)
			continue;
		if (divide_factor < 1) {
			// we would set the alarm higher than the current charge level.
			pDebug(DBG_DIAG, "divisor: %f, smaller than 1", divide_factor);
			goto ERROR;
		}
		alarm_value = (int)roundf(bd[x]->remaining_capacity / divide_factor);
		pDebug(DBG_DIAG, "Battery %d, remaining: %d, last full: %d, set alarm to: %d",
		       x, bd[x]->remaining_capacity, bd[x]->last_full_capacity, alarm_value);

		/*******  write alarm for each battery ************************/
		if (getDirEntry(x, dev_name, sizeof(dev_name), ACPI_BATTERY_DIR) < 0) {
			pDebug(DBG_DIAG, "could not get acpi directory name for battery %d", x);
			goto ERROR;
		}
		strncpy(file, ACPI_BATTERY_DIR, MAX_FILE_PATH);
		if (strlen(file) + strlen(dev_name) + strlen("/alarm" + 1) > MAX_FILE_PATH) {
			pDebug(DBG_DIAG, "Path %s/alarm too long", file);
			goto ERROR;
		} else {
			strcat(file, "/");
			strcat(file, dev_name);
			strcat(file, "/alarm");
			if (_write_line(file, "%d\n", alarm_value) < 0) {
				pDebug(DBG_INFO, "Could not write alarm value");
				goto ERROR;
			} else {
				/* don't set return value has already been set to 0 or initialised with 1 */
/* ##################### please review me ################################################ */
// can this work? and what fp are we using if this fails?
				if ((fp = fopen(file, "r")) == NULL) {
					pDebug(DBG_DIAG, "Could not read alarm settings made");
				}
/* ##################### please review me ################################################ */

				getColonValue(fp, verify, sizeof(verify), temp, sizeof(temp));
				if (atoi(verify) != alarm_value) {
					pDebug(DBG_DIAG, "failure. wrote: %d, read: %s", alarm_value, verify);
					fclose(fp);
					goto ERROR;
				}
				fclose(fp);
				pDebug(DBG_DEBUG, "Alarm value %d successfully written to %s", alarm_value, file);
			}
		}
		/*******  write alarm for each battery ************************/
	}
	for (x = 0; x < MAX_BATTERIES; x++) {
		if (bd[x] != NULL)
			free(bd[x]);
	}
	return ret;

      ERROR:
	for (x = 0; x < MAX_BATTERIES; x++) {
		if (bd[x] != NULL)
			free(bd[x]);
	}
	return UNKNOWN;

}
