/***************************************************************************
 *                                                                         *
 *                         Powersave Daemon                                *
 *                                                                         *
 *            Copyright (C) 2005 SUSE Linux Products GmbH                  *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2 of the License, or (at you   *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it will be useful, but     *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA                  *
 *                                                                         *
 ***************************************************************************/

#include "config.h"

#include "powersave_hal.h"
#include "powerlib_local.h"
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

#define ACPI_BATTERY_DIR  "/proc/acpi/battery"
#define ACPI_AC_DIR "/proc/acpi/ac_adapter"

extern DBusError dbus_error;
extern LibHalContext *hal_ctx;

int getBatteryDetailedInfo(const int no, BatteryDetailed * bd)
{
 	char *value;

	if (!ps_hal_init()) {
		/* error message already thrown in initHal() */
		return -1;
	}

	if (bd == NULL) {
		errno = EINVAL;
		return -3;
	}

	if (checkACPI() != ACPI) {
		return NO_ACPI_ERROR;
	}

	if (no > MAX_BATTERIES - 1 || no < 0) {
		return -2;
	}

	/* check whether module has been loaded and device is available */
	int res = check_ACPI_dir(ACPI_BATTERY_DIR);
	if (res < 0)
		/* returns either NO_DEVICE_ERROR or NO_MODULE_ERROR */
		return res;

	char **batteries;
	int numBatteries;
	batteries = libhal_find_device_by_capability(hal_ctx, "battery", &numBatteries, &dbus_error);

	if (dbus_error_is_set(&dbus_error)) {
		pDebug(DBG_INFO, "Could not get list of batteries in system: %s", dbus_error.message);
		ps_hal_free();
		dbus_error_free(&dbus_error);
		libhal_free_string_array(batteries);
		return 0;
	}

	/* if there are no batteries */
	if (numBatteries == 0) {
		libhal_free_string_array(batteries);
		return NO_DEVICE_ERROR;
	}

	/* if battery does not exist */
	if (no > numBatteries - 1 || no < 0) {
		libhal_free_string_array(batteries);
		return -2;
	}

 	/* check if found battery is a primary source (e.g. no ups) */
	value = ps_hal_get_property_string(batteries[no], "battery.type");
	if (value != NULL) {	
		if (strcmp(value, "primary") != 0) {
			libhal_free_string_array(batteries);
			libhal_free_string(value);
			return NO_DEVICE_ERROR;
		}
	}

	int present = ps_hal_get_property_bool(batteries[no],
					       "battery.present");

	if (present < 0)
		bd->present = UNKNOWN;
	/* if not present initialise everything with unknown and return! */
	else {
		if (present) {
			bd->present = PRESENT_YES;
		} else {
			bd->present = PRESENT_NO;
			bd->capacity_state = UNKNOWN;
			bd->charging_state = CHARG_STATE_UNKNOWN;
			bd->design_capacity_low = UNKNOWN;
			bd->design_voltage = UNKNOWN;
			bd->battery_technology = UNKNOWN;
			bd->last_full_capacity = UNKNOWN;
			bd->design_capacity = UNKNOWN;
			bd->present_voltage = UNKNOWN;
			bd->power_unit = UNKNOWN;
			bd->remaining_capacity = UNKNOWN;
			bd->present_rate = UNKNOWN;
			bd->capacity_granularity_2 = UNKNOWN;
			bd->capacity_granularity_1 = UNKNOWN;
			bd->design_capacity_warning = UNKNOWN;
			strcpy(bd->OEM_info, "UNKNOWN");
			strcpy(bd->serial_number, "UNKNOWN");
			strcpy(bd->model_number, "UNKNOWN");
			strcpy(bd->battery_type, "UNKNOWN");
			bd->alarm_limit = UNKNOWN;
			pDebug(DBG_INFO, "Battery %s not present", batteries[no]);
			libhal_free_string_array(batteries);
			return 0;
		}
	}

	/* capacity state */
	value = ps_hal_get_property_string(batteries[no],
					   "battery.charge_level.capacity_state");
	if (value != NULL) {
		if (!strcmp(value, "ok"))
			bd->capacity_state = STATE_OK;
		else if (!strcmp(value, "critical"))
			bd->capacity_state = STATE_CRIT;
		else {
			bd->capacity_state = UNKNOWN;
		}
		libhal_free_string(value);
	}

	/* --- get charging state --- */
	bd->charging_state = ps_hal_get_property_bool(batteries[no],
						      "battery.rechargeable.is_charging");

	if (bd->charging_state == 1)
		bd->charging_state = CHARG_STATE_CHARGING;
	else if (bd->charging_state == 0)
		bd->charging_state = CHARG_STATE_DISCHARGING;
	else
		bd->charging_state = CHARG_STATE_UNKNOWN;

	/* present_rate */
	bd->present_rate = ps_hal_get_property_int(batteries[no], "battery.charge_level.rate");
	if (bd->present_rate <= 0) {
		bd->present_rate = UNKNOWN;
	}

	/* --- get remaining capacity --- */
	bd->remaining_capacity = ps_hal_get_property_int(batteries[no],
							 "battery.charge_level.current");
	if (bd->remaining_capacity < 0) {
		bd->remaining_capacity = UNKNOWN;
	}

	/* get power unit */
	value = ps_hal_get_property_string(batteries[no], "battery.charge_level.unit");
	if (value != NULL) {
		if (!strcmp(value, "mWh")) {
			bd->power_unit = WATT_H;
		} else if (!strcmp(value, "mA")) {
			bd->power_unit = AMP;
		} else {
			bd->power_unit = UNKNOWN;
		}
		libhal_free_string(value);
	}

	/* present_voltage */
	bd->present_voltage = ps_hal_get_property_int(batteries[0],
						      "battery.charge_voltage.current");
	if (bd->present_voltage < 0) {
		bd->present_voltage = UNKNOWN;
	}

	/* design_capacity */
	bd->design_capacity = ps_hal_get_property_int(batteries[no],
						      "battery.charge_level.design");
	if (bd->design_capacity < 0) {
		bd->design_capacity = UNKNOWN;
	}

	/* last_full_capacity */
	bd->last_full_capacity = ps_hal_get_property_int(batteries[no],
							 "battery.charge_level.last_full");
	if (bd->design_capacity < 0) {
		bd->last_full_capacity = UNKNOWN;
	}

	/* battery_technology */
	bd->battery_technology = ps_hal_get_property_bool(batteries[no], "battery.is_rechargeable");
	if (bd->battery_technology) {
		bd->battery_technology = TECH_RECHARGEABLE;
	} else if (!bd->battery_technology) {
		bd->battery_technology = TECH_NON_RECHARGEABLE;
	} else {
		bd->battery_technology = UNKNOWN;
	}

	/* desing voltage */
	bd->design_voltage = ps_hal_get_property_int(batteries[no], "battery.voltage.design");
	if (bd->design_voltage < 0) {
		bd->design_voltage = UNKNOWN;
	}

	/* design_capacity_warning */
	bd->design_capacity_warning = ps_hal_get_property_int(batteries[no], "battery.charge_level.warning");
	if (bd->design_capacity_warning < 0) {
		bd->design_capacity_warning = UNKNOWN;
	}

	/* design_capacity_low */
	bd->design_capacity_low = ps_hal_get_property_int(batteries[no], "battery.charge_level.low");
	if (bd->design_capacity_low < 0) {
		bd->design_capacity_low = UNKNOWN;
	}

	/* capacity_granularity_1 */
	bd->capacity_granularity_1 = ps_hal_get_property_int(batteries[no], "battery.charge_level.granularity_1");
	if (bd->capacity_granularity_1 < 0) {
		bd->capacity_granularity_1 = UNKNOWN;
	}

	/* capacity_granularity_1 */
	bd->capacity_granularity_2 = ps_hal_get_property_int(batteries[no],
							     "battery.charge_level.granularity_2");
	if (bd->capacity_granularity_2 <= 0) {
		bd->capacity_granularity_2 = UNKNOWN;
	}

	/* alarm limit */
	bd->alarm_limit = ps_hal_get_property_int(batteries[no], "battery.alarm.design");
	if (bd->alarm_limit <= 0) {
		bd->alarm_limit = UNKNOWN;
	}

	value = ps_hal_get_property_string(batteries[no], "battery.model");
	if (value != NULL) {
		strcpy(bd->model_number, value);
		libhal_free_string(value);
	}
	else {
		strcpy(bd->model_number, "UNKNOWN");
	}

	value = ps_hal_get_property_string(batteries[no], "battery.serial");
	if (value != NULL) {
		strcpy(bd->serial_number, value);
		libhal_free_string(value);
	}
	else {
		strcpy(bd->serial_number, "UNKNOWN");
	}

	value = ps_hal_get_property_string(batteries[no], "battery.technology");
	if (value != NULL) {
		strcpy(bd->battery_type, value);
		libhal_free_string(value);
	}
	else {
		strcpy(bd->battery_type, "UNKNOWN");
	}

	value = ps_hal_get_property_string(batteries[no], "battery.vendor");
	if (value != NULL) {
		strcpy(bd->OEM_info, value);
		libhal_free_string(value);
	}
	else {
		strcpy(bd->OEM_info, "UNKNOWN");
	}

	libhal_free_string_array(batteries);

	return 0;
}

int getACAdapterStatus(void)
{
	int test;
	int ret;

	if (!ps_hal_init()) {
		/* error message already thrown in initHal() */
		return -1;
	}

	char **ac_adapter;
	char *ac_device;
	int num;

	test = checkACPI();
	/* --------- ACPI ------------- // */

	if (test == ACPI) {
		/* TODO: Only check for module here, because device
		 * availability is checked for above */
		ret = check_ACPI_dir(ACPI_AC_DIR);
		if (ret < 0)
			/* returns either NO_DEVICE_ERROR or NO_MODULE_ERROR */
			return ret;
	} else if (test == NOT_SUPPORTED) {
		pDebug(DBG_DIAG, "Neither APM nor ACPI support found");
		return AC_UNKNOWN;
	}

	ac_adapter = libhal_find_device_by_capability(hal_ctx, "ac_adapter", &num, &dbus_error);
	
	if (dbus_error_is_set(&dbus_error)) {
		pDebug(DBG_ERR, "Could not get ac_adapter device: %s", dbus_error.message);
		ps_hal_free();
		dbus_error_free(&dbus_error);
		libhal_free_string_array(ac_adapter);
		
		return NO_DEVICE_ERROR;
	}
	
	if (num > 0) {
		ac_device = strdup(*ac_adapter);
		libhal_free_string_array(ac_adapter);
	} else {
		return NO_DEVICE_ERROR;
	}
	
	int ac_state = ps_hal_get_property_bool(ac_device,
						"ac_adapter.present");
	free(ac_device);

	if (ac_state) {
		return AC_ONLINE;
	} else if (!ac_state) {
		return AC_OFFLINE;
	} else {
		return AC_UNKNOWN;
	}
	pDebug(DBG_INFO, "AC status and info read out");
	return ret;
}
